/**********************************************************************************************************************
 * DISCLAIMER
 * This software is supplied by Renesas Electronics Corporation and is only intended for use with Renesas products. No
 * other uses are authorized. This software is owned by Renesas Electronics Corporation and is protected under all
 * applicable laws, including copyright laws.
 * THIS SOFTWARE IS PROVIDED "AS IS" AND RENESAS MAKES NO WARRANTIES REGARDING
 * THIS SOFTWARE, WHETHER EXPRESS, IMPLIED OR STATUTORY, INCLUDING BUT NOT LIMITED TO WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. ALL SUCH WARRANTIES ARE EXPRESSLY DISCLAIMED. TO THE MAXIMUM
 * EXTENT PERMITTED NOT PROHIBITED BY LAW, NEITHER RENESAS ELECTRONICS CORPORATION NOR ANY OF ITS AFFILIATED COMPANIES
 * SHALL BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES FOR ANY REASON RELATED TO
 * THIS SOFTWARE, EVEN IF RENESAS OR ITS AFFILIATES HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
 * Renesas reserves the right, without notice, to make changes to this software and to discontinue the availability of
 * this software. By using this software, you agree to the additional terms and conditions found by accessing the
 * following link:
 * http://www.renesas.com/disclaimer
 *
 * Copyright (C) 2020 Renesas Electronics Corporation. All rights reserved.
 *********************************************************************************************************************/
/***********************************************************************/
/*                                                                     */
/*  FILE        :Main.c or Main.cpp                                    */
/*  DATE        :Tue, Oct 31, 2006                                     */
/*  DESCRIPTION :Main Program                                          */
/*  CPU TYPE    :                                                      */
/*                                                                     */
/*  NOTE:THIS IS A TYPICAL EXAMPLE.                                    */
/*                                                                     */
/***********************************************************************/
//#include "typedefine.h"
#include <machine.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mathf.h>
#include <complex.h>
#include "iodefine.h"
#include "ada_if.h"
#include "sci.h"
#include "board.h"
#include "sound_effect.h"
#include "r_dsp_typedefs.h"
#include "r_dsp_types.h"
#include "r_dsp_transform.h"
#include "r_dsp_complex.h"


#ifdef __cplusplus
//#include <ios>                        // Remove the comment when you use ios
//_SINT ios_base::Init::init_cnt;       // Remove the comment when you use ios
#endif

void main(void);
#ifdef __cplusplus
extern "C" {
void abort(void);
}
#endif

void main(void)
{
	unsigned char xc;
	unsigned long local_sw_state;
	
	int n;
	int offset = 0;
	int i, j, k, m;
	float a;
	
	float w[1024];
	
	int ret;
	
	const int data_handling_size = 1024;		//1024f[^FFT̏PʂƂi1/48kHz=20.83us@20.83us x 1024 = 21.33msj
	
	volatile unsigned long x;
	
	// FFT
	float x0[1024];
	float x1[1024];
	cplxf32_t X0[512];
	cplxf32_t X1[512];
	vector_t x0_vector = {1024, (void *)x0};
	vector_t x1_vector = {1024, (void *)x1};
	vector_t X0_vector = {512, (void *)X0};
	vector_t X1_vector = {512, (void *)X1};
	cplxf32_t fft_twiddles1[768]; // 32:24, 64:48, 128:96, 256:192, 512:384, 1024:768, 2048:1536
	uint32_t fft_bitrev1[240]; // 32:6, 64:12, 128:28, 256:56, 512:120, 1024:240, 2048:496
	r_dsp_fft_t h1 = {1024, 0, fft_twiddles1, fft_bitrev1, NULL, NULL};
	
	// amplitude spectrum
	float B[512];
	vector_t B_vector = {512, (void *)B};

	// IFFT
	float y[1024];
	cplxf32_t Y[512];
	vector_t y_vector = {1024, (void *)y};
	vector_t Y_vector = {512, (void *)Y};
	cplxf32_t fft_twiddles2[768]; // 32:24, 64:48, 128:96, 256:192, 512:384, 1024:768, 2048:1536
	uint32_t fft_bitrev2[240]; // 32:6, 64:12, 128:28, 256:56, 512:120, 1024:240, 2048:496
	r_dsp_fft_t h2 = {1024, 0, fft_twiddles2, fft_bitrev2, NULL, NULL};

	R_DSP_FFT_Init_f32cf32(&h1);
	R_DSP_FFT_Init_cf32f32(&h2);

	// Hanning window
	for (n = 0; n < 1024; n++)
	{
		w[n] = 0.5 - 0.5 * cosf(2.0 * M_PI * n / 1024.0);
	}
	
#if (USE_SCI == 1)
	//SCIgp
	sci_init();
	
	sci_write_str("\n");
	sci_write_str("Copyright (C) 2021 HokutoDenshi. All Rights Reserved.\n");
	sci_write_str("ADA-IF sample program (for HSBRX72M176, HSBRX72N176, HSBRX66N176, HSBRX72N144, HSBRX66N144).\n\n");
	sci_write_str("FFT/IFFT Vocorder.\n\n");
	
	sci_write_str("COMMAND USAGE:\n");
	sci_write_str("  SW2: Operation mode change\n");
	sci_write_str("  0-9: Operation mode change\n");
	sci_write_str("  p: print current status\n");
	
	sci_write_str("\n");
	sci_write_str("\nOperation mode -> Vocorder\n");
	
	sci_write_str("\n>");
#endif

	sci_write_flush();

	g_input_index = 0;								//̓obt@0珈Jn
	g_output_index = 0 - 1536 + DATA_BUFFER_SIZE;	//o̓obt@1536|CgOi1/48kHz=20.83us 20.83us x 256 = 32msjf[^i[n߂

	ada_if_init( );
	ret = ada_if_start( );
	
	if(ret != 0)
	{
		//{[hڑG[
#if (USE_SCI == 1)
		sci_write_str("\n\n");
		sci_write_str("ADA-IF board cannot initialize.\n\n");
		sci_write_str("Please check board connection.\n\n");
		sci_write_str("Program opeation stopped.\n");
#endif
		while(1)
		{
			__nop();
		}
	}
	
	local_sw_state = g_sw_state;
	
	while(1)
	{
		
		//f[^
		if(ada_if_input_index_diff(offset, g_input_index) > data_handling_size)	//̃f[^ data_handling_size ȏł΃f[^iDATA_BUFFER_SIZEdata_handling_size̐{Ƃj
		{
			switch(local_sw_state)
			{
				case 0:
					//{R[_
					for (n = 0; n < data_handling_size; n++)
					{
						m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						x0[n] = g_pcm0[m].ch0 * w[n];
					}

					for (n = 0; n < data_handling_size; n++)
					{
						m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						x1[n] = g_pcm0[m].ch1;
					}

					for (n = 0; n < data_handling_size; n++)
					{
						m = offset + n - 1;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						if (m < 0)
						{
							m += DATA_BUFFER_SIZE;
						}
						x1[n] -= 0.98 * g_pcm0[m].ch1;
						x1[n] *= w[n];
					}

					// FFT x[0-1024](real)) -> X[0-512](complex)
					R_DSP_FFT_f32cf32(&h1, &x0_vector, &X0_vector);
					R_DSP_FFT_f32cf32(&h1, &x1_vector, &X1_vector);

					// amplitude spectrum
					R_DSP_VecCplxMag_cf32f32(&X1_vector, &B_vector, 512);
					for (i = 0; i < 64; i++)
					{
						j = 8 * i;	
						a = 0.0;
						for(k = 0; k < 8; k++)
						{
							a += B[j + k];
						}
						a /= 8.0;
						for (k = 0; k < 8; k++)
						{
							B[j + k] = a;
						}
					}
					B[0] = 0.0;

					// DFT filter
					for (k = 0; k < (data_handling_size/2); k++)
					{
						Y[k].re = X0[k].re * B[k];
						Y[k].im = X0[k].im * B[k];
					}

					// IFFT Y[0-512](complex) -> y[0-1024](real)
					R_DSP_IFFT_CCS_cf32f32(&h2, &Y_vector, &y_vector);

					for (n = 0; n < (data_handling_size/2); n++)	//0-511f[^OiO񏈗㔼f[^add, I[obvAhj
					{
						m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						g_pcm1[m].ch0 += y[n];
						g_pcm1[m].ch1 += y[n];
					}
					
					for (n = (data_handling_size/2); n < data_handling_size; n++)	//512-1023f[^㔼iVKɃobt@Ɋi[j
					{
				         m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						g_pcm1[m].ch0 = y[n];
						g_pcm1[m].ch1 = y[n];
					}

					break;
					
				case 1:
					//X[injÔ݁j
					for (n = 0; n < data_handling_size; n++)		//0-1023f[^̃njOZ
					{
						m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						x0[n] = g_pcm0[m].ch0 * w[n];
						x1[n] = g_pcm0[m].ch1 * w[n];
					}

					for (n = 0; n < (data_handling_size/2); n++)	//0-511f[^OiO񏈗㔼f[^addj
					{
						m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						g_pcm1[m].ch0 += x0[n];
						g_pcm1[m].ch1 += x1[n];
					}
		
					for (n = (data_handling_size/2); n < data_handling_size; n++)	//512-1023f[^㔼iVKɃobt@Ɋi[j
					{
						m = offset + n;
						if (m >= DATA_BUFFER_SIZE)
						{
							m -= DATA_BUFFER_SIZE;
						}
						g_pcm1[m].ch0 = x0[n];
						g_pcm1[m].ch1 = x1[n];
					}

					break;
					
				
				default:
					break;
			}
			
			offset += (data_handling_size/2);		//f[^CfbNX͔(512)i߂
			if(offset >= DATA_BUFFER_SIZE)
			{
				offset -= DATA_BUFFER_SIZE;
			}
		}
		
		if(local_sw_state != g_sw_state)
		{
			local_sw_state = g_sw_state;
			
			//XCb`ꂽ
			switch(local_sw_state)
			{
				case 0:
					//{R[_
					sci_write_str("\nOperation mode(0) -> Vocorder\n");
					g_led_pattern = 0xffffffff;		//LED_
					break;
					
				case 1:
					//X[
					sci_write_str("\nOperation mode(1) -> Thru\n");
					g_led_pattern = 0x0;			//LED
					break;

			}
			sci_write_str("\n>");
		}
		
#if (USE_SCI == 1)
		xc = sci_read_char();
	
		if(xc != 0xff)
		{
			switch(xc)
			{
				case '0':
					if(g_sw_state != 0) sci_write_char(xc);
					g_sw_state = 0;
					break;
				
				case '1':
					if(g_sw_state != 1) sci_write_char(xc);
					g_sw_state = 1;
					break;

				case 'p':
					sci_write_char(xc);
					sci_write_str("\nvolume(0, 1, 2) = (");
					sci_write_uint16(g_ad_val[0]);
					sci_write_str(", ");
					sci_write_uint16(g_ad_val[1]);
					sci_write_str(", ");
					sci_write_uint16(g_ad_val[2]);
					sci_write_str(") sw = ");
					sci_write_uint32(g_sw_state);
					sci_write_str("\n\n>");
					break;
			}
		}
	
#endif

	}

}

#ifdef __cplusplus
void abort(void)
{

}
#endif
